# $Id: StochasticDAGGenerator.py,v 1.5 2007/07/28 17:26:39 kipp Exp $

#
# Preamble.
#

from glue import segments, lal
import onasys.DAGGenerator
from onasys import S5DAGGeneratorA
import logging
import popen2
import os, sys

#
# A DAG generator plug-in for making a single DAG that spans multiple
# segments.
#
# How to use this.  This DAG generator plug-in has a [daggenerator] section
# in the .ini file whose contents are identical to the [daggenerator]
# section used by the S5DAGGeneratorA plug-in.
#
# How this works.  This is based off of S5DAGGeneratorA.  The Config class
# for this plug-in is simply a copy of that plug-in's Config class (thus
# the .ini file is the same).  That plug-in defines a
# DAGGeneratorParameters class describing a DAG for a single segment
# (actually, just the start and stop times for the segment).  We define a
# DAGGeneratorParameters class for *this* plug-in as a list of
# DAGGeneratorParameters objects from the S5DAGGeneratorA plug-in.  We call
# that plug-in to generate its list, and then stick the entire list into
# one of our DAGGeneratorParameters objects.  We return that object as the
# single object in our list of DAGs returned to the daemon.
#

class DAGGeneratorConfig(S5DAGGeneratorA.DAGGeneratorConfig):
	def parse(self):
		for key in ["data_dir", "extra_args", "generator", "mkdagparamfile"]:
			self.__dict__[key] = self.parser.get(self.section, key)
		for key in ["cushion_timeout", "data_length_minimum", "analysis_length_minimum", "analysis_length_increment", "transient_left_minimum", "transient_left_increment", "transient_right_minimum", "transient_right_increment"]:
			self.__dict__[key] = self.parser.getfloat(self.section, key)

class DAGGeneratorParameters(onasys.DAGGenerator.DAGGeneratorParameters):
	def __init__(self, list):
		self.list = list
		self.extent = segments.segmentlist([dag.dataseg for dag in list]).extent()
                # list = list of segpairss; segpair = (dataseg, analseg)
                # want analysis boundaries
                self.analsegs = segments.segmentlist([dag.analseg for dag in list]).coalesce()
		self.datasegs = segments.segmentlist([dag.dataseg for dag in list]).coalesce()

	def __repr__(self):
		return repr(self.list)

def DAGGeneratorParameterList(clock, config, available, used):
	# Ask S5DAGGeneratorA to make a list of DAGs.
	list = S5DAGGeneratorA.DAGGeneratorParameterList(clock, config, available, used)

	# Put that list into our DAGGeneratorParameters object and return
	# a list containing that one object back to the daemon.  If empty,
	# return an empty list rather than a list containing only an empty list.
	if list == []:
		return []
	return [DAGGeneratorParameters(list)]

class DAGGenerator(onasys.DAGGenerator.DAGGenerator):
	def dag_name(self):
		return self.tag + "-" + str(int(self.params.extent[0])) + "-" + str(int(abs(self.params.extent))) + ".dag"

	def generate(self):
                newlyused = segments.segmentlist()

                try:
                        newlyused |= self._generate()
                except Exception, e:
                        logging.info("generator returned: %s" % str(newlyused))
			raise
		
                if len(newlyused) > 0:
                        if newlyused.extent() not in self.params.analsegs:
                                logging.warning("analysis segment to remove exceeds bounds of " + str(self.params.analsegs))
                else:
                        logging.warning("generator script returned no analysis segment for removal!")
		
                return {"analysis": newlyused}
	
	def _generate(self):
		""" This is the real worker code. """
		## Get parameters ##
		params = self._readConfig()
		params.update({'dagfilename': self.dag_name(),\
			       'path': self.config.data_dir,\
			       'dataseglist': self.params,\
			       'analseglist': self.params.analsegs})
		params['LSCdataFindServer'] = os.environ['LSC_DATAFIND_SERVER']
		
		# Fix paths; Use the -t argument to absolutize every file location
		tofix = ['jobfile','paramfile','outputdir','outputprefix','postDir','submitfilename','postprocsubmitfilename']
		for key in tofix:
			params[key] = os.path.join(params['path'],params[key])			
		## Add jobs to jobfile (1 start end duration)
		startjobnum = int(os.popen('wc -l %(jobfile)s' \
		                           % params).read().split()[0])+1
		lines = ["1 %d %d %d\n" % (x[0],x[1],x[1]-x[0])\
		         for x in self.params.datasegs]
		jobfileAddenda = "".join(lines)
		file(params['jobfile'],'a').write(jobfileAddenda)
		endjobnum = int(os.popen('wc -l %(jobfile)s' % params).read().split()[0])
		
		# Determine job numbers to submit - preformat as string
		params['jobnums'] = " ".join(map(str,range(startjobnum,endjobnum+1)))
		
		## Call stochastic_pipe.tclsh with the appropriate arguments ##
		# + is concatenate, so += is append
		callContents = " ".join(\
		    ["%(stochastic_pipe.tclsh)s", \
		#     "--force",\
		     "--jobsfile %(jobfile)s",\
		     "--jobList %(jobnums)s",\
		     "--dagfile %(dagfilename)s",\
		     "--paramsfile %(paramfile)s",\
		     "--submitfile %(submitfilename)s",\
		     "--server %(LSCdataFindServer)s",\
		     "--add-post-processing",\
		     "--postprocfile %(postprocsubmitfilename)s",\
		     "--postDir %(postDir)s",\
		     "--dSigmaCut %(dSigmaCut)s",\
		     "--largeSigmaCutoff %(largeSigmaCutoff)s",\
		     "--doRenormalize false",\
		     ])
		if params['verbose']==True:
		    callContents += " --verbose"
		logging.info("Calling %s\n" % (callContents % params))
		
		# Make the call, grabbing user input (empty) and output; log it.
		pipe = popen2.Popen4(callContents % params)
		pipe_output = pipe.fromchild.read()
		pipe_status = pipe.wait()
		logging.info("** stochastic_pipe.tclsh output **\n")
		logging.info(pipe_output)
		logging.info("** end stochastic_pipe.tclsh output **\n\n")
		
		if pipe_status != 0:
		    msg = "** stochstic_pipe.tclsh failed to create cache. Aborting.\n"
		    logging.error(msg)
		    raise onasys.DAGGenerator.DAGGeneratorError, msg
		return params['analseglist']


	def _readConfig(self):
		"""
		Read in configuration from file.
		Returns a dictionary of values.
		"""
		configfile = os.path.join(self.config.data_dir, \
                                          self.config.mkdagparamfile)
		import ConfigParser	
		config = ConfigParser.SafeConfigParser()
		config.optionxform = str  # make case-sensitive
		config.read(configfile)
		params = dict(config.items('mkdag'))
		params['verbose'] = bool(eval(params['verbose']))
		return params
