# $Id: onasysd.py,v 1.122 2007/03/09 14:53:31 kipp Exp $
#
# Copyright (C) 2005  Kipp C. Cannon
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Library of things useful to the daemon and segment and DAG generator plug-ins.
"""

#
# Preamble
#

import ConfigParser
from dagdbUpdator import BOSS
from glue import lal
from glue import LDRdataFindClient
from glue import LSCsegFindClient
from glue.gpstime import GpsSecondsFromPyUTC
from glue import segments
from glue import segmentsUtils
import libjobdb
import logging
import logging.handlers
import os
import pickle
import popen2
import re
from stat import S_ISDIR
from tempfile import mkstemp
import time

#
# Store this file's version number
#

__author = "Kipp Cannon <kipp@gravity.phys.uwm.edu>"
__version__ = "$Revision: 1.122 $"[11:-2]
__date__ = "$Date: 2007/03/09 14:53:31 $"[7:-2]


#
# Base class for all exceptions generated by onasysd library code
#

class OnasysError(Exception):
	"""
	Base class for exceptions raised by onasysd library code.
	"""
	pass


#
# Base class for configuration file interfaces
#

class ConfigError(OnasysError):
	"""
	Base class for exceptions raised by configuration parsers.
	"""
	pass

def ConfigErrorMsg(config, msg):
	"""
	Module level function for generating consistently-formated error
	messages related to configuration parsing.  config should be an
	instance of a Config object (or subclass), and msg is a text string
	to be used as the error message.
	"""
	return "[" + config.section + "]: " + msg

class Config(object):
	"""
	Base class for use in writing parsers for sections of the daemon's
	configuration file.  The section class variable contains the name
	of the section that is to be parsed.  A minimal subclass can be
	written by overriding only the section variable.  Typically, a
	subclass will provide overrides for the section variable, and the
	defaults(), parse() and validate() methods.
	"""

	section = "DEFAULTS"

	def __init__(self, filename):
		"""
		Construct a Config instance, reading input from the file
		named filename.
		"""
		self.filename = filename
		self.parser = ConfigParser.SafeConfigParser()
		try:
			self.parser.add_section(self.section)
		except ConfigParser.DuplicateSectionError:
			pass
		for default in self.defaults():
			self.parser.set(self.section, *default)
		self.parser.read(self.filename)

	def defaults(self):
		"""
		Returns a list of tuples containing key/value pairs as
		strings.  These are used as default values during parsing.
		"""
		return []

	def parse(self):
		"""
		Populates the class' attribute dictionary with the contents
		of the config file.  The values are stored as strings.
		"""
		for (name, value) in self.parser.items(self.section):
			self.__dict__[name] = value

	def validate(self):
		"""
		Verfies the correctness of the configuration loaded by the
		parse() method.  If the configuration is invalid, a
		ConfigError exception is raised.
		"""
		pass

	def log(self):
		"""
		Returns a list of strings to be used as a record of the
		configuration that has been loaded.
		"""
		list = []
		for (name, value) in self.parser.items(self.section):
			list.append(name + " = " + value)
		list.sort()
		list.insert(0, "[" + self.section + "]")
		return list


#
# onasysd configuration file interface
#

class OnasysdConfig(Config):
	"""
	Class implementing a parser for the [onasysd] section of the
	configuration file.
	"""

	section = "onasysd"

	def defaults(self):
		return [
			("dag_throttle", "5"),
			("daggenerator", "onasys.S5DAGGeneratorA"),
			("dryrun", "True"),
			("max_jobs_per_dag", "40"),
			("max_pre_per_dag", "1"),
			("max_post_per_dag", "1"),
			("pid_dir", "/tmp"),
			("purge_age_range", "604800:"),
			("purge_id_range", ":"),
			("purge_states", "success"),
			("purge_enable", "True"),
			("seggenerator", "onasys.LSCsegFindSegGenerator"),
			("send_mail", "False"),
			("state_file", "onasysd.%(tag)s.state"),
			("update_interval", "3600"),
			("virtual_start", "0")
		]

	def parse(self):
		self.dag_throttle = self.parser.getint(self.section, "dag_throttle")
		self.daggenerator = self.parser.get(self.section, "daggenerator")
		self.dryrun = self.parser.getboolean(self.section, "dryrun")
		self.e_mail = self.parser.get(self.section, "e-mail")
		self.pid_dir = self.parser.get(self.section, "pid_dir")
		try:
			self.purge_age_range = libjobdb.ParseRanges(self.parser.get(self.section, "purge_age_range").split(","))
		except ValueError, e:
			raise ConfigError, ConfigErrorMsg(self, "invalid purge_age_range: %s" % str(e))
		try:
			self.purge_id_range = libjobdb.ParseRanges(self.parser.get(self.section, "purge_id_range").split(","))
		except ValueError, e:
			raise ConfigError, ConfigErrorMsg(self, "invalid purge_id_range: %s" % str(e))
		self.purge_states = map(str.strip, self.parser.get(self.section, "purge_states").split(","))
		self.purge_enable = self.parser.getboolean(self.section, "purge_enable")
		self.seggenerator = self.parser.get(self.section, "seggenerator")
		self.send_mail = self.parser.getboolean(self.section, "send_mail")
		self.state_file = self.parser.get(self.section, "state_file")
		self.tag = self.parser.get(self.section, "tag")
		self.update_interval = self.parser.getint(self.section, "update_interval")
		self.virtual_start = lal.LIGOTimeGPS(self.parser.get(self.section, "virtual_start"))
		self.max_jobs_per_dag = self.parser.getint(self.section, "max_jobs_per_dag")
		self.max_pre_per_dag = self.parser.getint(self.section, "max_pre_per_dag")
		self.max_post_per_dag = self.parser.getint(self.section, "max_post_per_dag")

	def validate(self):
		if len(self.tag) > 30:
			raise ConfigError, ConfigErrorMsg(self, "tag length exceeds 30 characters")
		if re.search(r"\W", self.tag):
			raise ConfigError, ConfigErrorMsg(self, "tag must consist exclusively of alphanumeric characters and '_'")
		if self.pid_dir:
			if not S_ISDIR(os.stat(self.pid_dir).st_mode):
				raise ConfigError, ConfigErrorMsg(self, "pid_dir is not a directory")
			if not os.access(self.pid_dir, os.W_OK):
				raise ConfigError, ConfigErrorMsg(self, "pid_dir cannot be written to")
		if self.send_mail and not len(self.e_mail):
			raise ConfigError, ConfigErrorMsg(self, "e-mail notification requested, but no e-mail address supplied")
		if self.dag_throttle < 1 or self.dag_throttle > 50:
			raise ConfigError, ConfigErrorMsg(self, "dag_throttle must be in the range [1, 50]")
		if self.max_pre_per_dag < 1 or self.max_pre_per_dag > 10:
			raise ConfigError, ConfigErrorMsg(self, "max_pre_per_dag must be in the range [1, 10]")
		if self.max_post_per_dag < 1 or self.max_post_per_dag > 10:
			raise ConfigError, ConfigErrorMsg(self, "max_post_per_dag must be in the range [1, 10]")
		try:
			map(libjobdb.StateQuery, self.purge_states)
		except ValueError, e:
			raise ConfigError, ConfigErrorMsg(self, "invalid purge state %s" % str(e))


#
# Log interface
#

class LogFile(object):
	"""
	Class implementing an interface to a log file, with optional
	e-mailing of error messages.
	"""

	def __init__(self, filename):
		"""
		Construct a log file instance that will write output to the
		file named filename.
		"""
		self.log = logging.getLogger()
		# first remove, flush, and close pre-existing handlers
		for handler in self.log.handlers[:]:
			self.log.removeHandler(handler)
			handler.flush()
			handler.close()
		# add new log file handler
		self.log.setLevel(logging.NOTSET)
		handler = logging.FileHandler(filename)
		handler.setFormatter(logging.Formatter("%(asctime)s: %(name)s (%(process)d): %(levelname)s: %(message)s"))
		self.log.addHandler(handler)

	def add_email(self, addr, fromaddr = os.environ["LOGNAME"] + "@" + os.popen("hostname -f").read().rstrip(), subject = ""):
		"""
		Add an e-mail address to which error messages are to be
		mailed.  A new address is added each time this method is
		called().
		"""
		[toaddr, mailhost] = addr.split("@")
		handler = logging.handlers.SMTPHandler(mailhost, fromaddr, toaddr, subject)
		handler.setFormatter(logging.Formatter("%(asctime)s: %(name)s (%(process)d): %(levelname)s: %(message)s"))
		handler.setLevel(logging.ERROR)
		self.log.addHandler(handler)


#
# Saved state interface
#

class StateFile(object):
	"""
	Class implementing an interface to a file used for preserving
	internal state information.
	"""

	def __init__(self, filename):
		"""
		Construct an instance of this class, using the file named
		filename to store the state object.
		"""
		self.set_filename(filename)

	def set_filename(self, filename):
		"""
		Set the name of the file used for storing state
		information.
		"""
		self.filename = str(filename)

	def save(self, obj):
		"""
		Save the Python object obj in the saved-state file,
		overwriting the current contents.
		"""
		pickle.dump(obj, file(self.filename, "w"))

	def restore(self):
		"""
		Extract and return the Python object currently stored in
		the saved-state file.
		"""
		return pickle.load(file(self.filename, "r"))


#
# Time calculations.
#

class Clock(object):
	"""
	Class implementing a time-keeping device with the ability to have
	an offset imposed based on a choice of start time.  The offset is
	computed based on the actual time at which the time keeping device
	was started, and a simulated or virtual start time.
	"""

	def __init__(self):
		self.start = lal.LIGOTimeGPS(0)
		self.virtual_start = lal.LIGOTimeGPS(0)

	def set_start(self, start):
		"""
		Set the time at which the clock should be marked as having
		started.  The difference between this time and the current
		time is added to the "virtual start" time to arrive at the
		effective current time when applying an offset to the
		clock.
		"""
		self.start = start

	def set_virtual_start(self, virtual_start):
		"""
		Set the clock's simulated or virtual start time.
		"""
		self.virtual_start = virtual_start

	def nownow(self):
		"""
		The (real) current time according to the system clock.
		"""
		return lal.LIGOTimeGPS(GpsSecondsFromPyUTC(time.time()))

	def now(self):
		"""
		The simulated current time.  This is the same as the real
		current time (the output of the nownow() method) unless a
		virtual start time has been selected, in which case it is
		shifted by the difference between the virtual start time
		and the actual start time.
		"""
		if self.virtual_start:
			return self.nownow() - self.start + self.virtual_start
		return self.nownow()

	def maxend(self):
		"""
		A time for use as an upper bound in queries to the segfind
		and datafind servers, and so on.  This is the simulated
		current time plus an additional padding.
		"""
		if self.virtual_start:
			return self.now()
		return self.nownow() + 60


#
# Fractal length generation.
#

class FractalInterval(object):
	"""
	Class defining an iterable object which returns numbers in the
	range (0, 1] in a fractal pattern.  The algorithm is to keep track
	of an integer counter that is incremented by 1 upon each call to
	the next() method, and then return as a fraction of the total count
	the value of the most significant bit that changed.  The pattern
	that results is

	1/1, 2/2, 1/3, 4/4, 1/5, 2/6, 1/7, 8/8, 1/9, 2/10, 1/11, 4/12,
	1/13, 2/14, 1/15, 16/16, ...
	"""

	def __init__(self):
		self.count = 0

	def __iter__(self):
		return self

	def next(self):
		self.count += 1
		length = self.count ^ (self.count - 1)
		length ^= length >> 1
		return float(length) / self.count


#
# The window in which to query for data availability.
#

class SearchWindow(FractalInterval):
	"""
	A class defining a time window within which servers should be
	queried for things.
	"""

	def __init__(self):
		FractalInterval.__init__(self)
		self.window = segments.segment(-segments.infinity(), +segments.infinity())

	def next(self, lower, upper, minlength):
		if upper - lower < minlength:
			self.window = segments.segment(lower, upper)
		else:
			start = min(self.window[1], upper - minlength)
			start -= FractalInterval.next(self) * (start - lower)
			self.window = segments.segment(start, upper)
		return self.window


#
# PID file interface
#

class PIDFile(object):
	"""
	Class defining the interface to the daemon's .pid and .info files.
	"""

	def __init__(self, base, pid, info, dir = ""):
		"""
		Create and initialize the .pid and .info files.  The text
		string base is used as a tag in the file names, the process
		ID pid is written into the .pid file, and the Python object
		info is pickled and written into the .info file.  The text
		string dir can be used to choose a location for the files;
		the default is /tmp.
		"""
		self.base = str(base)
		self.pid = pid
		self.pid_filename = ""
		self.info = StateFile(None)
		if str(dir):
			(f, self.pid_filename) = mkstemp(".pid", "onasysd." + self.base + ".", str(dir))
		else:
			(f, self.pid_filename) = mkstemp(".pid", "onasysd." + self.base + ".")
		os.write(f, str(self.pid) + "\n")
		os.close(f)
		self.info.set_filename(self.pid_filename[:-3] + "info")
		self.set_info(info)

	def set_info(self, info):
		"""
		Save the Python object info in the .info file.
		"""
		self.info.save(info)

	def touch(self):
		"""
		Update the time stamps on both the .pid and .info files.
		"""
		for name in [self.pid_filename, self.info.filename]:
			try:
				os.utime(name, None)
			except:
				pass

	def __del__(self):
		"""
		Delete the .pid and .info files when we are deleted.
		"""
		for name in [self.pid_filename, self.info.filename]:
			try:
				os.remove(name)
			except:
				pass


#
# Job scheduler interface
#

class SchedulerError(OnasysError):
	"""
	Class for exceptions generated by the interface to the scheduler.
	"""
	pass

class Scheduler(object):
	"""
	Class defining an interface to the job scheduling system.
	"""

	def __init__(self, max_jobs_per_dag, max_pre_per_dag, max_post_per_dag):
		self.load_query = libjobdb.UserQuery(os.environ["LOGNAME"])
		self.load_query &= libjobdb.StateQuery("wait") | libjobdb.StateQuery("idle") | libjobdb.StateQuery("run")
		self.max_jobs_per_dag = max_jobs_per_dag
		self.max_pre_per_dag = max_pre_per_dag
		self.max_post_per_dag = max_post_per_dag

	def load_command(self):
		"""
		Return the command executed to query the load of the
		scheduler.
		"""
		return "boss_q --state wait,idle,run --user " + os.environ["LOGNAME"]

	def get_load(self):
		"""
		Query the scheduler's load.  The result is cached in this
		class' load attribute.
		"""
		logging.info(self.load_command())
		try:
			return BOSS.BossJob.select(self.load_query).count()
		except Exception, e:
			raise SchedulerError, str(e)

	def submit_command(self, dag):
		"""
		Return the command used to submit the DAG named dag.
		"""
		return "boss_submit_dag --max-jobs %d --max-pre %d --max-post %d %s" % (self.max_jobs_per_dag, self.max_pre_per_dag, self.max_post_per_dag, str(dag))

	def submit(self, dag):
		"""
		Submit the DAG file named dag.
		"""
		logging.info(self.submit_command(dag))
		child = popen2.Popen3(self.submit_command(dag), True)
		for line in child.childerr:
			logging.warning(line.rstrip())
		for line in child.fromchild:
			logging.info(line.rstrip())
		status = child.wait()
		if not os.WIFEXITED(status) or os.WEXITSTATUS(status):
			raise SchedulerError, "scheduler submission returned " + str(os.WEXITSTATUS(status))


#
# DAG working directory constructor
#

class WorkingDirectory(object):
	"""
	Class defining an iterable object for sequentially constructing
	working directories in which DAGs are created and executed.
	"""

	def __init__(self, template):
		"""
		Construct a new WorkingDirectory object using the text
		string template as the format for the directory names.
		template should be of the form

		"blah/blah/blah/000/0/000/0/..."

		that is, it should be a path starting with some depth of
		directories followed by a depth of numeric directories.
		The part of the template consisting of numeric directories
		must begin with a leading /, so if no higher non-numeric
		directories are desired then template should begin with
		"./".

		The numeric directories are used to initialize the integer
		directory counter, which will be used to generate new
		directories in the sequence.
		"""
		try:
			i = re.search("(/[-0-9]+)+/?$", template).start()
		except AttributeError:
			raise ValueError, template
		self.base = template[:i]
		template = template[i:].split("/")
		self.n = long("".join(template))
		self.breaks = map(len, template)
		for i in range(1, len(self.breaks)):
			self.breaks[i] += self.breaks[i-1]
		self.breaks.reverse()
		self.format = "%0" + str(self.breaks[0]) + "d"
		self.max = 10L**self.breaks[0] - 1

	def __str__(self):
		"""
		Return the directory name constructed from the original
		template and the current value of the integer directory
		counter.
		"""
		s = list(self.format % self.n)
		for i in self.breaks[1:-1]:
			s[i:i] = "/"
		return os.path.join(self.base, "".join(s))

	def next(self):
		"""
		Increment the directory counter by 1.
		"""
		self.n += 1
		if self.n > self.max:
			raise StopIteration
		return self

	def __iter__(self):
		return self

def FindWorkingDirectory(path):
	"""
	Module level function that attempts to construct a WorkingDirectory
	template by recursively searching below path for numeric
	directories.  The output is path with the highest numbered numeric
	subdirectories found beneath it appended.
	"""
	while True:
		l = os.listdir(path)
		l.sort()
		try:
			int(l[-1])
		except:
			return path
		path = os.path.join(path, l[-1])

def ListWorkingDirectories(path):
	"""
	Returns a list of all potential WorkingDirectory paths below path.
	This is done by recursively searching below path for numeric
	subdirectories.  No guarantee is made that all entries in the
	resulting list could have been constructed from the same template.
	"""
	subdirs = os.listdir(path)
	if not len(subdirs):
		return [path]
	try:
		map(int, subdirs)
	except ValueError:
		return [path]
	subdirs.sort()
	return reduce(list.__add__, map(lambda d: ListWorkingDirectories(os.path.join(path, d)), subdirs), [])


def RemoveWorkingDirectory(path, verbose = None):
	"""
	Deletes the WorkingDirectory path.  If verbose is a callable
	object, it will be passed a single string argument for generating
	"verbosity" messages during the removal process.
	"""
	# check that path exists and is a directory
	try:
		if not S_ISDIR(os.stat(path).st_mode):
			raise ValueError, "%s: not a directory" % path
	except OSError, e:
		if e.errno == 2:
			raise ValueError, "%s: no such file or directory" % path
		# something else is wrong
		raise e

	# check that the path can be parsed as a working directory
	try:
		WorkingDirectory(path).base
	except ValueError:
		raise ValueError, "%s not a valid working directory" % path

	# remove the working directory's contents
	if verbose:
		verbose("deleting contents of %s" % path)
	for root, dirs, files in os.walk(path, topdown=False):
		for name in files:
			os.remove(os.path.join(root, name))
		for name in dirs:
			if os.path.islink(os.path.join(root, name)):
				os.remove(os.path.join(root, name))
			else:
				os.rmdir(os.path.join(root, name))

	# remove the working directory
	os.rmdir(path)
	if verbose:
		verbose("%s deleted" % path)


def PruneWorkingDirectories(path, verbose = None):
	"""
	Recursively deletes empty numeric directories below path If verbose
	is a callable object, it will be passed a single string argument
	for generating "verbosity" messages during the pruning process.
	"""
	# remove empty numeric child directories
	# FIXME: this is broken:  it deletes any empty numeric directory,
	# whether or not it consitutes part of a valid working directory tree.
	def _recurse(path, verbose):
		entries = os.listdir(path)
		for dir in entries:
			try:
				int(dir)
			except ValueError:
				# name not an integer;  since this
				# directory contains non-integer entries,
				# it cannot be part of the job directory
				# hierarchy so bail out indicating this
				# entry was not deleted
				return False
			if _recurse(os.path.join(path, dir), verbose):
				# entry was deleted
				entries.remove(dir)
		if not entries:
			# this is now an empty numeric directory
			try:
				os.rmdir(path)
				# deleted
				return True
			except Exception, e:
				if verbose:
					verbose("skipping %s: %s" % (path, str(e)))
		# not deleted
		return False

	if verbose:
		verbose("pruning empty directories below %s" % path)
	for dir in os.listdir(path):
		try:
			int(dir)
		except ValueError:
			# name not an integer, so path isn't part of the
			# job directory hiearchy, so bail out
			return
		_recurse(os.path.join(path, dir), verbose)


#
# LSCsegFind interface
#

class SegFindConfig(Config):
	"""
	Class implementing a parser for a [segfind] section in the
	configuration file.
	"""

	section = "segfind"

	def defaults(self):
		try:
			return [("server", os.environ["LSC_SEGFIND_SERVER"])]
		except KeyError:
			return []

	def parse(self):
		self.instrument = self.parser.get(self.section, "instrument")
		self.min_query_length = self.parser.getfloat(self.section, "min_query_length")
		self.set_server(self.parser.get(self.section, "server"))
		self.type = self.parser.get(self.section, "type")

	def set_server(self, server):
		self.host = str(server)
		try:
			[self.host, self.port] = self.host.split(":")
			self.port = int(self.port)
		except:
			self.port = None

class SegFind(object):
	"""
	Class implementing an interface to an LSCsegFind server.
	"""

	def __init__(self, config):
		"""
		Construct a SegFind object using the SegFindConfig object
		config for configuration information.
		"""
		self.config = config

	def command(self, bounds):
		"""
		Return an LSCsegFind client command line equivalent to the
		query performed by a call to the list() method.  This is
		useful for placing in log files so that queries can be
		easily reproduced by hand if needed.
		"""
		command = "LSCsegFind --type " + self.config.type + " --interferometer " + self.config.instrument + " --gps-start-time " + str(int(bounds[0])) + " --gps-end-time " + str(int(bounds[1]))
		if self.config.host:
			command += " --server " + self.config.host
			if self.config.port:
				command += ":" + str(self.config.port)
		return command

	def list(self, bounds):
		"""
		Query for segments in the interval defined by the
		segments.segment object bounds.  The result is returned as
		a segmentlist object.
		"""
		if self.config.port:
			client = LSCsegFindClient.LSCsegFind(self.config.host, self.config.port)
		else:
			client = LSCsegFindClient.LSCsegFind(self.config.host)
		list = client.findStateSegments({"interferometer" : self.config.instrument, "type" : self.config.type, "start" : str(int(bounds[0])), "end" : str(int(bounds[1])), "lfns" : False, "strict" : True})
		return segments.segmentlist([segments.segment(*map(lal.LIGOTimeGPS, seg)) for seg in list])


#
# LSCdataFind interface
#

class DataFindConfig(Config):
	"""
	Class implementing a parser for a [datafind] section in the
	configuration file.
	"""

	section = "datafind"

	def defaults(self):
		list = [
			("lfn_only", "False"),
			("match", ""),
			("padding", "3600"),
			("url_type", "")
		]
		try:
			list.append(("server", os.environ["LSC_DATAFIND_SERVER"]))
		except KeyError:
			pass
		return list

	def parse(self):
		self.lfn_only = self.parser.getboolean(self.section, "lfn_only")
		self.match = self.parser.get(self.section, "match")
		self.observatory = self.parser.get(self.section, "observatory")
		self.padding = lal.LIGOTimeGPS(self.parser.get(self.section, "padding"))
		self.set_server(self.parser.get(self.section, "server"))
		self.type = self.parser.get(self.section, "type")
		self.url_type = self.parser.get(self.section, "url_type")

	def set_server(self, server):
		self.host = str(server)
		try:
			[self.host, self.port] = self.host.split(":")
			self.port = int(self.port)
		except:
			self.port = None

	def validate(self):
		if self.lfn_only and (len(self.match) or len(self.url_type)):
			raise ConfigError, ConfigErrorMsg(self, "cannot specify match or url_type when searching for LFNs only")

class DataFind(object):
	"""
	Class defining an interface to an LSCdataFind server.
	"""

	def __init__(self, config):
		"""
		Construct a DataFind object using the DataFindConfig object
		config for configuration information.
		"""
		self.config = config

	def command(self, bounds):
		"""
		Return an LSCdataFind client command line equivalent to the
		query performed by a call to the list() method.  This is
		useful for placing in log files so that queries can be
		easily reproduced by hand if needed.
		"""
		command = "LSCdataFind --observatory " + self.config.observatory + " --type " + self.config.type + " --gps-start-time " + str(int(bounds[0] - self.config.padding)) + " --gps-end-time " + str(int(bounds[1]))
		if self.config.lfn_only:
			command += " --names-only"
		else:
			if self.config.url_type:
				command += " --url-type " + self.config.url_type
			if self.config.match:
				command += " --match " + self.config.match
		if self.config.host:
			command += " --server " + self.config.host
			if self.config.port:
				command += ":" + str(self.config.port)
		return command

	def list(self, bounds):
		"""
		Query for files in the interval defined by the
		segments.segment object bounds.  The result is returned as
		a segmentlist object representing the intervals spanned by
		the file list that was retrieved.
		"""
		if self.config.port:
			client = LDRdataFindClient.LSCdataFindClient(self.config.host, self.config.port)
		else:
			client = LDRdataFindClient.LSCdataFindClient(self.config.host)
		args = {
			"observatory" : self.config.observatory,
			"type" : self.config.type,
			"start" : str(int(bounds[0] - self.config.padding)),
			"end" : str(int(bounds[1])),
			"offset" : None,
			"limit" : None,
			"strict" : None
		}
		if self.config.lfn_only:
			query = client.findFrameNames
		elif self.config.match or self.config.url_type:
			query = client.findFrameURLsFilter
			args["match"] = self.config.match
			args["urlType"] = self.config.url_type
		else:
			query = client.findFrameURLs
		return segmentsUtils.fromfilenames(query(args), coltype=lal.LIGOTimeGPS).coalesce()


#
# A class providing automatic rounding of a segment to discrete lengths.
#

class StepSegment(object):
	"""
	A class providing rounding/truncating of segments to discrete
	lengths.  This is useful for writing DAG generators, since a common
	problem enountered is the need to adjust a segment to match certain
	criteria imposed by the analysis pipeline.

	In particular, this class describes segments having lengths given
	by (min + n inc) where n in a non-negative integer.
	"""
	def __init__(self, min, inc):
		"""
		Initialize a StepSegment class object.  Set the minimum
		segment length to min, and the length increment to inc.
		"""
		self.min = float(min)
		self.inc = float(inc)

	def truncate_start(self, seg):
		"""
		Tuncate the segment seg at the nearest allowed boundary by
		moving the start of the segment towards the end.
		"""
		if abs(seg) < self.min:
			return segments.segment(seg[1], seg[1])
		if not self.inc:
			return segments.segment(seg[1] - self.min, seg[1])
		return segments.segment(seg[0] + (abs(seg) - self.min) % self.inc, seg[1])

	def truncate_end(self, seg):
		"""
		Truncate the segment seg at the nearest allowed boundary by
		move the end of the segment towards the start.
		"""
		if abs(seg) < self.min:
			return segments.segment(seg[0], seg[0])
		if not self.inc:
			return segments.segment(seg[0], seg[0] + self.min)
		return segments.segment(seg[0], seg[1] - (abs(seg) - self.min) % self.inc)

	def add_start(self, seg, x):
		"""
		Increase the length of the segment by the largest amount
		not greater than x by moving the start of the segment to an
		allowed boundary.
		"""
		return self.truncate_start(segments.segment(seg[0] - x, seg[1]))

	def add_end(self, seg, x):
		"""
		Increase the length of the segment by the largest amount
		not greater than x by moving the end of the segment to an
		allowed boundary.
		"""
		return self.truncate_end(segments.segment(seg[0], seg[1] + x))
